# Copyright (c) 2008, Kiyoko F. Aoki-Kinoshita, Soka University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without 
# modification, are permitted provided that the following conditions are met:
#  - Redistributions of source code must retain the above copyright notice, 
#    this list of conditions and the following disclaimer.
#  - Redistributions in binary form must reproduce the above copyright notice, 
#    this list of conditions and the following disclaimer in the documentation 
#    and/or other materials provided with the distribution.
#  - Neither the name of Soka University nor the names of its contributors may 
#    be used to endorse or promote products derived from this software without 
#    specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
# IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
# OF THE POSSIBILITY OF SUCH DAMAGE.

######################################################################################
#   This is a program that calculates the prediction score of each glycan structure  #
# by using co-occurrence score and the glycosyltransferase expression data.          #
# Using the equation for the prediction score by Suga, glycans are output in order   #
# of high prediction score.                                                          #
######################################################################################

require "matrix"

t0 = Time.now

## method for making gnumH
def getRPfromGnum(gnum,file)
    seen = 0                               ## seen: flag for consecutive Gnumbers
    gnumH = Hash.new                       ## gnumH: hash containing information for each Gnumber, key = RP, value = #RPs
    file = open(ARGV[2])                   ## ARGV[2] <= name of file containing Gnumber, RP and #RP
    while line = file.gets
        if line.include?(gnum)
	    grnum = line.chomp!.split(/\t/)       ## grnum: grnum[0]=Gnumber, grnum[1]=RP, grnum[2]=#RP
	    gnumH[grnum[1]] = grnum[2].to_i
	    seen += 1
	elsif seen > 0
	    break
        end
    end
    file.close
    return gnumH
end

## rounding the score to the second decimal point
class Numeric
    def round_n(nth)
        score = self*(10**(-nth))
        return score.round()*(10**nth)
    end
end

## calculating sum of Array 
class Array
    def sum
        inject(nil) { |sum,x| sum ? sum+x : x}
    end
end

############ reading in gnum,rpArray and coScore #################################

if ARGV.size < 4 
    print "Usage: ruby calcPredScore.rb [co-occurrence score matrix] [expression data file] [gnumber-RP-#RPs file] [number of output scores]\n"
    exit;
end

gnum = Array.new                           ## gnum : Gnumber  
rpArray = Array.new                        ## rpArray : Reaction Pattern(=RP)
coScore = Array.new                        ## coScore : Co-occurrence score 
coMatrix = Array.new                       ## coMatrix : matrix of co-occurrence score
row = Array.new
counter = 1                                ## counter: number of co-occurrence scores read (used to monitor count of each row)
mode = 0                                   ## mode=0: Gnumber, mode=1: RP, mode=2: Co-occurrence score

filename = ARGV[0]                         ### ARGV[0] <= file name of Co-occurrence Score file generated by makeCoScore.rb
file = open(filename)
while grc = file.gets do                   ## grc: RP of Candidate Glycan
    if grc == "///\n"
        mode += 1
        next
    end

    if mode == 0
        gnum << grc.chomp!                 ## push Gnumber into gnum Array
    elsif mode == 1
        rpArray << grc.chomp.downcase     ## push RP into rpArray
    else 
        sCount = rpArray.size
	if counter < sCount
            row << grc.chomp!.to_f
            counter += 1
	else
            row << grc.chomp!.to_f
            counter = 1
            coScore << row
            row = Array.new
	end
    end
end
coMatrix = Matrix[*coScore]                ## coMatrix: co-occurrence score Matrix
file.close

################# Reading of query data ###########################
queH = Hash.new                            ## queH: key = RP, value = microarray expression data
filename = ARGV[1]                         ### ARGV[1] <= file of microarray expression data
file = open(filename)
que  = file.read
que1 = que.split("\n")
qcount = que1.size                         ## qcount: RP size of query

for i in 0..qcount-1
    query = que1[i].split("\t")

## push RP and microarray expression data of query into queH
    if queH.key?(query[0])           
        if queH[query[0]] < query[1]       ## query[0]: RP
           queH[query[0]] = query[1]       ## query[1]: microarray expression data
	end
    else
        queH[query[0]] = query[1]
    end
end
file.close
######################## Calculation of prediction score ##################################
sEH = Hash.new{|sc,gn| sc[gn] = []}        ## sEH:key = prediction score, value = Gnumber
number = ARGV[3].to_i                      ## number <=ARGV[3]: number of glycans to output as results 
exp = Array.new                            ## exp: microarray expression data
m =Array.new                               ## m: coMatrix column
count = Array.new                          ## count: size of gnumH for each Gnumber
sco2 = Array.new                           ## sco2: sum of co-occurrence score * expression values for candidate glycans 
h = 0                                      ## h: sEH value size

## prediction score equation = (1/count[i]) * sum(m * exp) 
## push co-occurrence score of RP of Query and Candidate into m
queH.each{|rp,ex|
    qri =  rpArray.index(rp.downcase)      ## qri:Query RP index

    if qri == nil
        next
    end
    exp << ex.to_f
    m << coMatrix.column(qri).to_a
}

ceM = Matrix[*m].transpose * Vector[*exp]  ## ceM: co-occurrence score * microarray expression values

## making of gnumH for gnum
gnum.each{|glycan|     
    gnumH = getRPfromGnum(glycan,ARGV[2])
      
    count << gnumH.size
    sco1 = Array.new                       ## sco1: working Array of co-occurrence score * expression values for glycan
    gnumH.each{|k,v|   
        cri =  rpArray.index(k.downcase)
        if cri == nil 
	    sco1 = nil
            break
        end
        sco1 << ceM[cri] 
    }
    if sco1 == nil
	sco2 << nil
    else
        sco2 << sco1.sum
    end
}

## Sum of ceM[i] divided by gnumH size
for i in 0..sco2.size-1
    if sco2[i] == nil
        next
    end
    sco = sco2[i] / count[i]
    score = sco.round_n(-2)
        
## store only the high-ranking prediction scores and Gnumbers in sEH
    if h < number
        sEH[score] << gnum[i]
        h += 1
    else
        smin = sEH.keys.min                ## smin: Minimum value of working prediction scores in sEH
        if score >= smin
	    sEH[score] << gnum[i]
            h += 1
            if sEH[score].size == 1
	        hd =sEH[smin].size      
	        h -= hd
	        sEH.delete_if{|key, value|key == smin}
            else
	        hd = sEH[smin].size
	        if h-hd >= number
		    sEH.delete_if{|key, value|key == smin }
		    h -= hd
	        end
            end
        end
    end
end
        
## sorting of prediction score
cnt = 1                                    ## cnt: count of prediction score
rank = 0 
sEH.sort_by{|k,v| -k.to_f}.each{|score,gs|
    rank = cnt
    gs.each{|g|
        print cnt,"\t",rank , "\t", score, "\t", g, "\n"
        cnt += 1
    }
}
t1 = Time.now
print "Time is ", t1 - t0,"\n"
