# Copyright (c) 2008, Kiyoko F. Aoki-Kinoshita, Soka University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without 
# modification, are permitted provided that the following conditions are met:
#  - Redistributions of source code must retain the above copyright notice, 
#    this list of conditions and the following disclaimer.
#  - Redistributions in binary form must reproduce the above copyright notice, 
#    this list of conditions and the following disclaimer in the documentation 
#    and/or other materials provided with the distribution.
#  - Neither the name of Soka University nor the names of its contributors may 
#    be used to endorse or promote products derived from this software without 
#    specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
# IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
# OF THE POSSIBILITY OF SUCH DAMAGE.

###############################################################################################
# This is a program that calculates co-occurrence score by using the correlation coefficient. #
# The correlation coefficient uses the Reaction Patern(=RP) of all glycan structures and the  #
# number of the RP.                                                                           #
# RP is the information of the disaccharide and its corresponding linkage information in the  #
# glycan structure.                                                                           #
###############################################################################################

## method for making gnumH 
def getRPfromGnum(gnum,file)
    seen = 0                               ## seen: kind of number of RP in each Gnumber
    gnumH = Hash.new                       ## gnumH: hash containing information for each Gnumber, key = RP, value = #RPs
    file.rewind
    while line = file.gets
        if line.include?(gnum)
	    grnum = line.chomp!.split(/\t/) ## grnum: grnum[0]=Gnumber, grnum[1]=RP, grnum[2]=#RP
	    gnumH[grnum[1]] = grnum[2].to_i
	    seen += 1
	elsif seen > 0
	    break
    	end
    end
    return gnumH
end

## method of sum of Array
class Array
    def sum
        inject(nil) { |sum,x| sum ? sum+x : x}
    end
end

############ making of table of Gnumbers and the number of specific RPs #############################

if ARGV.size < 3 
    print "Usage: ruby makeCoScore.rb [gnumber file] [gnumber-RP-#RPs file] [outfile]\n"
    exit;
end


gnum = Array.new                             ## gnum: Gnumber
rpArray = Array.new                          ## rpArray: Reaction Pattern(=RP) 
rpgArray = Array.new                         ## rpgArray: table of Gnumbers and #RPs
file = open(ARGV[0])                         ### ARGV[0] <= File of Gnumbers

## Reading Gnumber 
while gnum1 = file.gets do
    gnum << gnum1.chomp!
end
file.close

count = gnum.size                            ## count: number of Gnumbers

file = open(ARGV[1])                   ### ARGV[1] <= File of Gnumber, RP and #RP
## make table of Gnumbers and #RPs
gnum.each{|glycan| 
p glycan
    gnumH = getRPfromGnum(glycan,file) 
    gidx = gnum.index(glycan)                ## gidx: index of glycan in gnum array
    
    gnumH.each{|karray,v|

## make all RPs downcase for easier comparison
        k = karray.downcase                  ## k = downcased RP
        if k =~ /dol/ || k =~ /ser/ || k =~ /cer/ || k =~ /:[sp]$/ || k =~ /^[sp]:/ || k =~ /[\(\),\*\/]/ || k =~ /hex/ || k =~ /unknown/ || k =~ /-ol/
            next;
        end
        ridx =  rpArray.index(k)             ## ridx: index of k in the rpArray

## push k into rpArray if k is not in rpArray
        if ridx == nil         
            rpArray << k       
            ridx =  rpArray.index(k) 
        end

## (If the number of rpgArray.size is larger than that of ridx, a new Array is made)
## if rpgArray is full, new matrix to store Gnumbers and #RPs is made
        if ridx >= rpgArray.size
	    rpgArray << Array.new(count,0)   ## rpgArray: default=0
	end
        rpgArray[ridx][gidx] = v
    }
}
file.close

################ calculation of Co-occurrence Score ##############################
aveArray = Array.new
sCount = rpArray.size

## making of aveArray
for i in 0..sCount-1
    sum = 0
    rpg2Array = rpgArray[i]
    sum = rpg2Array.sum
    ave = sum.to_f / count.to_f              ## ave: Average of Number of RP of all Glycan
    aveArray << ave.to_f
end

coScore = Array.new(sCount) {Array.new(sCount,0)}  ## coScore: table of co-occurrence score

## 1.0 is put in the co-occurrence score of the same RP
for i in 0..sCount-1
    coScore[i][i] = 1.0
end

## Co-occurrence score is calculated by using correlation coefficient 
## score[i,j] = sum((rpg2Array[k] - aveArray[i]) * (rpg3Array[k] - aveArray[j]))/(sum(rpg2Array[k] - aveArray[i])^2 * sum (rpg3Array[k] - aveArray[j])^2)^(1/2)
for i in 0..sCount-1 
    rpg2Array = rpgArray[i]
    for j in i+1..sCount-1
        sum1 = 0  
        sum2 = 0 
        sum3 = 0 
        x = 0 
        y = 0 
        coScore1 = 0
        rpg3Array = rpgArray[j]

	#rpg2Array.collect! {|k| k-aveArray[i] }
	#rpg3Array.collect! {|k| k-aveArray[j] }

        for k in 0..count-1
            x = rpg2Array[k] - aveArray[i]
            y = rpg3Array[k] - aveArray[j]
            sum1 += x * y
            sum2 += x * x
            sum3 += y * y
        end

        coScore1 = sum1 / ((sum2 * sum3) ** 0.5)
    
        coScore[j][i] = coScore1       
        coScore[i][j] = coScore1
    end
end

############# save file into given filename ################## 
filename = ARGV[2]                           ### ARGV[2] <= file name to be saved
File.open(filename, "w+"){|file|
    file.puts(gnum , "///" , rpArray , "///" , coScore)
}
